#! There is no good nor evil but that which is created by man.

import constants
import pygame
from pygame import gfxdraw
import images
import mysprites
import potential
import vector
import data
from pgu import isovid

import random
btn_height = 90
btn_wid = 128

nope = pygame.mixer.Sound(data.filepath('nope.ogg'))

hugefont = pygame.font.Font(data.filepath("FreeMonoBold.ttf"), 130)

class Bumper(object):
  def __init__(self, message, duration, next):
    self.dur = duration
    self.gfx = hugefont.render(message, True, (255,0,0))
    self.tick = 0
    self.next = next
  def update(self, events):
    self.tick += 1
    if self.tick > self.dur:
      return True
  def render(self, surface):
    surface.fill((0,0,0))
    surface.blit(self.gfx,(10,250))
    pygame.display.update()

class TitleScreen(object):
  def __init__(self):
      pygame.mixer.music.load(data.filepath('bgm1.ogg'))
      pygame.mixer.music.set_volume(0.55)
      self.start = True
      self.state_changed = False
      self.gfx = images.title
  def update(self, events):
    if self.start:
      self.start = False
      pygame.mixer.music.play()
    for e in events:
      if (e.type is pygame.MOUSEBUTTONDOWN and e.button==1) or  e.type is pygame.KEYDOWN :
        self.state_changed = True
        pygame.mixer.music.fadeout(2000)
    return self.state_changed
  
  def render(self, surface):
    surface.blit(self.gfx,(0,0))

    pygame.display.update()

class GameEnv(object):
  walltiles = set([1])
  fontsize = 9
  bigfontsize = 18

  def check_place_valid(self, pos):
    if pos in self.placed_locs or pos in self.discopot.wallset:
      return False
    
    return True

  def place_pylon(self, pos):
    item = mysprites.Pylon(pos)
    self.iso_eng.sprites.append(item)

    self.discopot.wallset.add(pos)
    self.iter_solve = True
    self.placed_locs.add(pos)
    
  def place_bloom(self, pos):
    self.placed_locs.add(pos)
    t = mysprites.Boomtower(pos, 1, 3)
    self.iso_eng.sprites.append(t)
    self.boomers.append(t)
    
  def place_joni(self, pos):
    self.placed_locs.add(pos)
    t = mysprites.Boomtower(pos, 2, 4)
    self.iso_eng.sprites.append(t)
    self.boomers.append(t)

  def place_gordon(self, pos):
    self.placed_locs.add(pos)
    t = mysprites.Boomtower(pos, 3, 5)
    self.iso_eng.sprites.append(t)
    self.boomers.append(t)
  tower_ranges = {1:3, 2:4, 3:5}
  def __init__(self, surface, tile_image_path, level_image_path, start_view, player_info, total, quota, spawn_rate):
    self.button_def = [
             (self.place_pylon, (images.pylon[0][0], "Pylon", 50)),
             (self.place_bloom,(images.boomer[0][0], "Bloomberg News", 100)),
             (self.place_joni,(images.boomer[0][0], "Joni Mitchell", 300)),
             (self.place_gordon,(images.boomer[0][0], "Gordon Lightfoot", 300)),
             ]

    self.spawn_rate = spawn_rate
    self.player_info = player_info
    self.boomers = []
    self.ticks = 0
    self.iter_solve = False
    self.font = pygame.font.Font(data.filepath("FreeMonoBold.ttf"), self.fontsize)
    self.bigfont = pygame.font.Font(data.filepath("FreeMonoBold.ttf"), self.bigfontsize)

    self.surface = surface
    self.iso_eng = isovid.Isovid()
    self.iso_eng.screen = self.surface
    self.iso_eng.tga_load_tiles(tile_image_path,(constants.tile_width,constants.tile_height))
    self.iso_eng.tga_load_level(level_image_path, bg=1)
    self.map = self.iso_eng.tlayer
    self.center = start_view 
    self.cur_btn = 0
    #self.discoman = discoman
    #self.iso_eng.sprites.append(discoman)
    self.discopot = potential.PotentialMap(self.iso_eng.tlayer, self.iso_eng.clayer, self.walltiles, [1], [2])
    self.discopot.solve(300, 0.001)
    self.potimg = pygame.transform.scale2x(self.discopot.to_image())
    self.buttons = []
    self.mkhud()
    self.placed_locs = set()
    self.defunked = 0
    self.total = total
    self.quota = quota
    self.boogiers = 0
    self.state_changed = False
    self.get_ready = hugefont.render("GET READY", True, (255,0,0))
  def mkhud(self):

    def draw_btn(surf, img, label, cost):

      offs = (5, 5)
      pygame.draw.rect(surf, (0x80,0x80,0x80),pygame.Rect(5, 5, btn_wid - 10, btn_height - 10))
      surf.blit(img, offs)
      surf.blit(self.font.render(label, True, (0xff,0xff,0xff)), (5, 5))
      surf.blit(self.font.render("$%d" % cost, True, (0xff,0xff,0xff)), vector.addvec((5, 5), (0, btn_height - 20)))
    start_pos = (0,constants.screen_height - btn_height)
    for act, btn in self.button_def:
      btn_img = pygame.Surface((btn_wid, btn_height), pygame.SRCALPHA)

      draw_btn(btn_img, *btn)
      self.buttons.append((pygame.Rect(start_pos[0],start_pos[1],btn_wid,btn_height), btn_img))
      start_pos = vector.addvec(start_pos, (btn_wid + 5,0))

  def set_view_pos(self, vpos):
    self.iso_eng.view.x, self.iso_eng.view.y = vector.addvec(self.iso_eng.tile_to_view(vpos), (-400, -300))

  def draw_HUD(self):
    c = 0
    for rect, img in self.buttons:
      if c == self.cur_btn:
        color = (0xe0,0xe0,0x30)
      else:
        color = (0x30,0x30,0x30)
      pygame.draw.rect(self.surface, color,rect)
      self.surface.blit(img, rect) 
      c+=1
    clr = (80,255,255)
    money_img = self.bigfont.render("money:%d" % self.player_info.money, True, clr)
    defunked = self.bigfont.render("De-Funked: %d / %d" % (self.defunked, self.quota), True, clr)
    boogie_progress = self.bigfont.render("Boogie procession: %d" % (self.boogiers), True, clr)
    self.surface.blit(money_img, (10,10))
    self.surface.blit(defunked, (10,10 + self.bigfontsize))
    self.surface.blit(boogie_progress, (10,10 + 2*self.bigfontsize))
  
  def set_view(self, pos):
    self.iso_eng.view.x, self.iso_eng.view.y = self.iso_eng.tile_to_view(pos)
    self.iso_eng.view.x -= constants.screen_width /2
    self.iso_eng.view.y -= constants.screen_height /2

  def grok_input(self, events):
    for e in events:
      if e.type is pygame.MOUSEBUTTONDOWN and e.button==3:
        self.center = self.iso_eng.screen_to_tile(e.pos)
        self.set_view(self.center)
      if e.type is pygame.KEYDOWN:
        if e.key == pygame.K_UP:
          self.center = vector.addvec(self.center, (-1, -1))
        if e.key == pygame.K_DOWN:
          self.center = vector.addvec(self.center, (1, 1))
        if e.key == pygame.K_LEFT:
          self.center = vector.addvec(self.center, (-1, 1))
        if e.key == pygame.K_RIGHT:
          self.center = vector.addvec(self.center, (1, -1))
        if self.center[0] <0: self.center = (0, self.center[1])
        if self.center[0] >= self.discopot.width: self.center = (self.discopot.width, self.center[1])
        if self.center[1] <0: self.center = ( self.center[0], 0)
        if self.center[1] >= self.discopot.height: self.center = (self.center[0], self.discopot.height )
        self.set_view(self.center)
      if e.type is pygame.MOUSEBUTTONDOWN and e.button==1:
        b = 0
        mset = False
        for rect, img in self.buttons:
          if rect.collidepoint(e.pos):
            self.cur_btn = b
            mset = True
          b+= 1
        if not mset:
          act, etc = self.button_def[self.cur_btn]
          tile = self.iso_eng.screen_to_tile(e.pos)
          cost = etc[2]
          if self.check_place_valid(tile) and (self.player_info.money - cost>= 0):
            act(self.iso_eng.screen_to_tile(e.pos))
            self.player_info.money -= cost
          else:
            nope.play()


  def update(self, events):
    
    if self.iter_solve:
      change = self.discopot.solve(1, 0.0002)
      if change <= 0.00021:
        self.iter_solve = False

    self.grok_input(events)
    self.ticks += 1
    self.update_sprites()
    self.iso_eng.loop()
    return self.state_changed

    #self.iso_eng.view.x, self.iso_eng.view.y = vector.addvec((-400, -300), self.iso_eng.tile_to_view(self.discoman.pos))

  def update_sprites(self):
    if self.ticks % self.spawn_rate == 0 and self.boogiers < self.total and self.ticks > 120:
      spawnpos = random.sample(list(self.discopot.posset), 1)[0]
      spawnpos = vector.addvec(spawnpos, (random.random()*.3,random.random()*0.3))
      newsprite = mysprites.Disco(spawnpos)
      self.iso_eng.sprites.append(newsprite)
      self.boogiers += 1
    human_sprites = 0
    for disco in self.iso_eng.sprites:

      if disco.__class__ in (mysprites.Disco, mysprites.Square):
        human_sprites += 1

      ret = disco.update(self)
      if not ret:
        self.iso_eng.sprites.remove(disco)
    
    if self.boogiers == self.total and human_sprites == 0:
      self.state_changed = True
      if self.defunked > self.quota:
        self.endstate = "Win"
      else:
        self.endstate = "Lose"

  def highlight_mouse(self):
    pos = pygame.mouse.get_pos()
    tpos = self.iso_eng.screen_to_tile(pos)
    self.hightlight_tile(tpos)


  def hightlight_tile(self, tpos):
    p1 = self.iso_eng.tile_to_screen(tpos)
    p2 = self.iso_eng.tile_to_screen(vector.addvec(tpos, (1,0)))
    p3 = self.iso_eng.tile_to_screen(vector.addvec(tpos, (1,1)))
    p4 = self.iso_eng.tile_to_screen(vector.addvec(tpos, (0,1)))
    c= (255*(self.ticks%2),255*(self.ticks%2),255*(self.ticks%2))
    pygame.draw.lines(self.surface, c, True, [p1,p2,p3,p4])
    if self.cur_btn in (1,2,3):
      rng = self.tower_ranges[self.cur_btn]
      gfxdraw.ellipse(self.surface, p1[0],p1[1],constants.tile_width *rng, constants.tile_width * rng /2, c)
  def render(self, surf):
    self.surface.fill((0,0,0))

    self.iso_eng.update(self.surface)

    self.highlight_mouse()
    #self.potimg = pygame.transform.scale2x(self.discopot.to_image())
    #self.surface.blit(self.potimg, (0,0))
    self.draw_HUD()
    if self.ticks < 120:
      self.surface.blit(self.get_ready, (50,250))
    #for disco in self.iso_eng.sprites:
    #  pygame.draw.circle(self.surface, (255,0,0), self.iso_eng.tile_to_screen(disco.pos), 3)
    pygame.display.update()
