import math
import vector
import pygame
import copy
import time

# Blah Blah
# Laplace Solver
# Blah Blah

class PotentialMap(object):
  def __init__(self, tilemap, codemap, wallcodes, poscodes, negcodes):

    self.width = len(tilemap[0])
    self.height = len(tilemap)
    self.potmap  = [[0.0] * self.width for y in xrange(self.height)]
    self.wallset = self.getcells(tilemap, wallcodes)

    self.posset = self.getcells(codemap, poscodes)
    self.negset = self.getcells(codemap, negcodes)

  def getcells(self,cmap, codes):
    rset = set()
    for y in xrange(self.height):
      for x in xrange(self.width):
        if cmap[y][x] in codes:
          rset.add((x,y))
    return rset

  def getpot_int(self, pos):
    x,y = pos
    if (x < 0 or x >= self.width) or (y <0 or y >= self.height):
      return 0.0, 0.0
    if pos in self.wallset: return (0.0, 0.0)
    if pos in self.posset: return (1.0, 1.0)
    if pos in self.negset: return (1.0, 0)
    return (1.0, self.potmap[y][x])

  def is_wall(self, pos):
    return vector.vfloor(pos) in self.wallset

  def get_grad(self, pos):
    ix, iy = (int(math.floor(pos[0])), int(math.floor(pos[1])))

    if (ix, iy) in self.wallset:
      return 0.0, 0.0
    pmap = self.potmap
    use, pot = self.getpot_int((ix, iy))
    use, xpot1 = self.getpot_int((ix-1, iy))
    if not use: xpot1 = pot
    use, xpot2 = self.getpot_int((ix+1, iy))
    if not use: xpot2 = pot
    use, ypot1 = self.getpot_int((ix, iy-1))
    if not use: ypot1 = pot
    use, ypot2 = self.getpot_int((ix, iy+1))
    if not use: ypot2 = pot
    return (xpot1 - xpot2, ypot1 - ypot2)


  def is_const(self, pos):
    if (pos in self.wallset) or (pos in self.posset) or (pos in self.negset):
      return True

  def solve(self, iterations, max_change_cutoff):
    start_time = time.time()
    max_change = 0
    pmap = self.potmap
    for i in xrange(iterations):
      max_change = 0
      for y in xrange(self.height):
        for x in xrange(self.width):
          if self.is_const((x,y)):
            pmap[y][x] = self.getpot_int((x,y)) [1]
            continue 
          acc = self.getpot_int((x-1, y))
          acc = vector.addvec(acc, self.getpot_int((x+1, y)))
          acc = vector.addvec(acc, self.getpot_int((x, y-1)))
          acc = vector.addvec(acc, self.getpot_int((x, y+1)))
          edg, pot = acc
          if not edg:
            continue 
          new = pot / edg
          old = pmap[y][x]
          diff = new - old
          adiff = abs(diff)
          if adiff > max_change: max_change = adiff
          pmap[y][x] = old + diff * 1.9 # 1.9 acceleration factor. seems stable
      if max_change < max_change_cutoff:
        break
    #dur = time.time() - start_time
    return max_change
    #print "solved %d iterations in %f seconds:%f" % (i, dur, iterations / dur)
    #print consts
    #print self.potmap
  
  def to_image(self):
    surf = pygame.Surface((self.width, self.height))
    parray = pygame.PixelArray(surf)
    for y in xrange(self.height):
      for x in xrange(self.width):
        pot = int(300*self.potmap[y][x])

        if pot > 255:
          pot = 255
        if pot < 0:
          pot = 0
        parray[x][y] = (pot, pot , pot)
    del parray
    return surf